#include "atlas_asm.h"
/*
 * This file does a 12x8 unrolled mvt_sse with these params:
 *    CL=16, ORDER=clmajor
 */
#ifndef ATL_GAS_x8664
   #error "This kernel requires x86-64 assembly!"
#endif
/*
 * Integer register assignment
 */
#define M       %rdi
#define N       %rsi
#define pA0     %rdx
#define lda     %rax
#define pX      %r8
#define pY      %r9
#define II      %rbx
#define pX0     %r11
#define Mr      %rcx
#define incAXm  %r10
#define incII   %r15
#define incAn   %r14
#define lda3    %r12
#define lda5    %r13
#define lda7    %rbp
/*
 * SSE register assignment
 */
#define rA0     %xmm0
#define rX0     %xmm1
#define rY0     %xmm2
#define rY1     %xmm3
#define rY2     %xmm4
#define rY3     %xmm5
#define rY4     %xmm6
#define rY5     %xmm7
#define rY6     %xmm8
#define rY7     %xmm9

/*
 * macros
 */
#ifndef MOVA
   #define MOVA movaps
#endif
#define movapd movaps
#define movupd movups
#define xorpd xorps
#define addpd addps
#define mulpd mulps
#define addsd addss
#define mulsd mulss
#define movsd movss
#define haddpd haddps
/*
 * Define macros controlling prefetch
 */
#ifndef PFDIST
   #define PFDIST 256
#endif
#ifndef PFADIST
   #define PFADIST PFDIST
#endif
#ifndef PFYDIST
   #define PFYDIST 64
#endif
#ifndef PFXDIST
   #define PFXDIST 64
#endif
#ifndef PFIY
   #ifdef ATL_3DNow
      #define PFIY prefetchw
   #else
      #define PFIY prefetchnta
   #endif
#endif
#ifndef PFIX
   #define PFIX prefetcht0
#endif
#ifndef PFIA
   #define PFIA prefetchnta
#endif
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #define prefA(mem) PFIA mem
#endif
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #define prefY(mem) PFIY mem
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #define prefX(mem) PFIX mem
#endif
/*
 *                      %rdi        %rsi           %rdx          %rcx
 * void ATL_UGEMV(ATL_CINT M, ATL_CINT N, const TYPE *A, ATL_CINT lda,
 *                          %r8      %r9
 *                const TYPE *X, TYPE *Y)
 */
.text
.text
.global ATL_asmdecor(ATL_UGEMV)
ALIGN64
ATL_asmdecor(ATL_UGEMV):

/*
 * Save callee-saved iregs
 */
   movq %rbp, -8(%rsp)
   movq %rbx, -16(%rsp)
   movq %r12, -24(%rsp)
   movq %r13, -32(%rsp)
   movq %r14, -40(%rsp)
   movq %r15, -48(%rsp)
/*
 * Compute M = (M/MU)*MU, Mr = M - (M/MU)*MU
 * NOTE: Mr is %rcx reg, so we can use jcx to go to cleanup loop
 */
   mov  %rcx, lda       /* move lda to assigned register, rax */
   movl   $12*16, -56(%rsp)       /* mem = MU */
   fildl  -56(%rsp)                     /* ST = MU */
   movl   %edi, -56(%rsp)               /* mem = M */
   fidivrl -56(%rsp)                    /* ST = M/MU */
   fisttpl -60(%rsp)                    /* mem = TRUNC(M/MU) */
   movl    -60(%rsp), %ebx              /* rbx = TRUNC(M/MU) */
   imul    $12*16, %ebx, %ebx     /* rbx = MU*TRUNC(M/MU) */
   mov     M, Mr                        /* Mr = M */
   sub     %rbx, Mr                     /* Mr = M - MU*TRUNC(M/MU) */
   mov     %rbx, M                      /* M  = MU*TRUNC(M/MU) */
/*
 * Setup constants
 */
   mov lda, incAn       /* incAn = lda */
   sub M, incAn         /* incAn = lda - (M/MU)*MU */
   sub Mr, incAn        /* incAn = lda - M */
   shl $2, incAn        /* incAn = (lda-M)*sizeof */
   shl $2, lda          /* lda *= sizeof */
   sub $-128, pA0       /* code compaction by using signed 1-byte offsets */
   sub $-128, pX        /* code compaction by using signed 1-byte offsets */
   mov pX, pX0          /* save for restore after M loops */
   mov $-768, incAXm     /* code comp: use reg rather than constant */
   lea (lda, lda,2), lda3       /* lda3 = 3*lda */
   lea (lda, lda,4), lda5       /* lda5 = 5*lda */
   lea (lda3,lda,4), lda7       /* lda7 = 7*lda */
   add lda7, incAn              /* incAn = (8*lda-M)*sizeof */
   mov $16*12, incII      /* code comp: use reg rather than constant */
   mov M, II
   ALIGN32
   LOOPN:
      #ifdef BETA0
         xorpd rY0, rY0
         xorpd rY1, rY1
         xorpd rY2, rY2
         xorpd rY3, rY3
         xorpd rY4, rY4
         xorpd rY5, rY5
         xorpd rY6, rY6
         xorpd rY7, rY7
      #else
         movsd 0(pY), rY0
         movsd 4(pY), rY1
         movsd 8(pY), rY2
         movsd 12(pY), rY3
         movsd 16(pY), rY4
         movsd 20(pY), rY5
         movsd 24(pY), rY6
         movsd 28(pY), rY7
      #endif

      LOOPM:
         movapd 0-128(pX), rX0
         MOVA   0-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+0(pA0))

         MOVA   0-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+0(pA0,lda))
         MOVA   0-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+0(pA0,lda,2))
         MOVA   0-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+0(pA0,lda3))
         MOVA   0-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+0(pA0,lda,4))
         MOVA   0-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+0(pA0,lda5))
         MOVA   0-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+0(pA0,lda3,2))
         MOVA   0-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+0(pA0,lda7))

         movapd 64-128(pX), rX0
         MOVA   64-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+64(pA0))

         MOVA   64-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+64(pA0,lda))
         MOVA   64-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+64(pA0,lda,2))
         MOVA   64-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+64(pA0,lda3))
         MOVA   64-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+64(pA0,lda,4))
         MOVA   64-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+64(pA0,lda5))
         MOVA   64-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+64(pA0,lda3,2))
         MOVA   64-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+64(pA0,lda7))

         movapd 128-128(pX), rX0
         MOVA   128-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+128(pA0))

         MOVA   128-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+128(pA0,lda))
         MOVA   128-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+128(pA0,lda,2))
         MOVA   128-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+128(pA0,lda3))
         MOVA   128-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+128(pA0,lda,4))
         MOVA   128-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+128(pA0,lda5))
         MOVA   128-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+128(pA0,lda3,2))
         MOVA   128-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+128(pA0,lda7))

         movapd 192-128(pX), rX0
         MOVA   192-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+192(pA0))

         MOVA   192-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+192(pA0,lda))
         MOVA   192-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+192(pA0,lda,2))
         MOVA   192-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+192(pA0,lda3))
         MOVA   192-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+192(pA0,lda,4))
         MOVA   192-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+192(pA0,lda5))
         MOVA   192-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+192(pA0,lda3,2))
         MOVA   192-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+192(pA0,lda7))

         movapd 256-128(pX), rX0
         MOVA   256-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+256(pA0))

         MOVA   256-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+256(pA0,lda))
         MOVA   256-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+256(pA0,lda,2))
         MOVA   256-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+256(pA0,lda3))
         MOVA   256-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+256(pA0,lda,4))
         MOVA   256-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+256(pA0,lda5))
         MOVA   256-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+256(pA0,lda3,2))
         MOVA   256-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+256(pA0,lda7))

         movapd 320-128(pX), rX0
         MOVA   320-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+320(pA0))

         MOVA   320-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+320(pA0,lda))
         MOVA   320-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+320(pA0,lda,2))
         MOVA   320-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+320(pA0,lda3))
         MOVA   320-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+320(pA0,lda,4))
         MOVA   320-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+320(pA0,lda5))
         MOVA   320-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+320(pA0,lda3,2))
         MOVA   320-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+320(pA0,lda7))

         movapd 384-128(pX), rX0
         MOVA   384-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+384(pA0))

         MOVA   384-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+384(pA0,lda))
         MOVA   384-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+384(pA0,lda,2))
         MOVA   384-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+384(pA0,lda3))
         MOVA   384-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+384(pA0,lda,4))
         MOVA   384-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+384(pA0,lda5))
         MOVA   384-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+384(pA0,lda3,2))
         MOVA   384-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+384(pA0,lda7))

         movapd 448-128(pX), rX0
         MOVA   448-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+448(pA0))

         MOVA   448-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+448(pA0,lda))
         MOVA   448-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+448(pA0,lda,2))
         MOVA   448-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+448(pA0,lda3))
         MOVA   448-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+448(pA0,lda,4))
         MOVA   448-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+448(pA0,lda5))
         MOVA   448-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+448(pA0,lda3,2))
         MOVA   448-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+448(pA0,lda7))

         movapd 512-128(pX), rX0
         MOVA   512-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+512(pA0))

         MOVA   512-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+512(pA0,lda))
         MOVA   512-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+512(pA0,lda,2))
         MOVA   512-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+512(pA0,lda3))
         MOVA   512-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+512(pA0,lda,4))
         MOVA   512-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+512(pA0,lda5))
         MOVA   512-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+512(pA0,lda3,2))
         MOVA   512-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+512(pA0,lda7))

         movapd 576-128(pX), rX0
         MOVA   576-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+576(pA0))

         MOVA   576-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+576(pA0,lda))
         MOVA   576-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+576(pA0,lda,2))
         MOVA   576-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+576(pA0,lda3))
         MOVA   576-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+576(pA0,lda,4))
         MOVA   576-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+576(pA0,lda5))
         MOVA   576-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+576(pA0,lda3,2))
         MOVA   576-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+576(pA0,lda7))

         movapd 640-128(pX), rX0
         MOVA   640-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+640(pA0))

         MOVA   640-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+640(pA0,lda))
         MOVA   640-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+640(pA0,lda,2))
         MOVA   640-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+640(pA0,lda3))
         MOVA   640-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+640(pA0,lda,4))
         MOVA   640-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+640(pA0,lda5))
         MOVA   640-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+640(pA0,lda3,2))
         MOVA   640-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+640(pA0,lda7))

         movapd 704-128(pX), rX0
         MOVA   704-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0
         prefA(PFADIST+704(pA0))

         MOVA   704-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         prefA(PFADIST+704(pA0,lda))
         MOVA   704-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         prefA(PFADIST+704(pA0,lda,2))
         MOVA   704-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         prefA(PFADIST+704(pA0,lda3))
         MOVA   704-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         prefA(PFADIST+704(pA0,lda,4))
         MOVA   704-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         prefA(PFADIST+704(pA0,lda5))
         MOVA   704-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         prefA(PFADIST+704(pA0,lda3,2))
         MOVA   704-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7
         prefA(PFADIST+704(pA0,lda7))

         movapd 16-128(pX), rX0
         MOVA   16-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   16-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   16-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   16-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   16-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   16-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   16-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   16-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 32-128(pX), rX0
         MOVA   32-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   32-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   32-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   32-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   32-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   32-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   32-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   32-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 48-128(pX), rX0
         MOVA   48-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   48-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   48-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   48-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   48-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   48-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   48-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   48-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 80-128(pX), rX0
         MOVA   80-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   80-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   80-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   80-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   80-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   80-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   80-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   80-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 96-128(pX), rX0
         MOVA   96-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   96-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   96-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   96-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   96-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   96-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   96-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   96-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 112-128(pX), rX0
         MOVA   112-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   112-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   112-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   112-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   112-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   112-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   112-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   112-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 144-128(pX), rX0
         MOVA   144-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   144-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   144-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   144-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   144-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   144-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   144-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   144-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 160-128(pX), rX0
         MOVA   160-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   160-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   160-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   160-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   160-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   160-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   160-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   160-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 176-128(pX), rX0
         MOVA   176-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   176-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   176-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   176-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   176-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   176-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   176-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   176-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 208-128(pX), rX0
         MOVA   208-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   208-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   208-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   208-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   208-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   208-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   208-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   208-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 224-128(pX), rX0
         MOVA   224-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   224-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   224-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   224-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   224-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   224-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   224-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   224-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 240-128(pX), rX0
         MOVA   240-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   240-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   240-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   240-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   240-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   240-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   240-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   240-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 272-128(pX), rX0
         MOVA   272-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   272-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   272-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   272-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   272-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   272-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   272-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   272-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 288-128(pX), rX0
         MOVA   288-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   288-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   288-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   288-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   288-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   288-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   288-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   288-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 304-128(pX), rX0
         MOVA   304-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   304-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   304-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   304-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   304-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   304-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   304-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   304-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 336-128(pX), rX0
         MOVA   336-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   336-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   336-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   336-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   336-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   336-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   336-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   336-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 352-128(pX), rX0
         MOVA   352-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   352-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   352-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   352-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   352-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   352-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   352-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   352-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 368-128(pX), rX0
         MOVA   368-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   368-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   368-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   368-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   368-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   368-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   368-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   368-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 400-128(pX), rX0
         MOVA   400-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   400-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   400-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   400-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   400-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   400-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   400-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   400-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 416-128(pX), rX0
         MOVA   416-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   416-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   416-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   416-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   416-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   416-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   416-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   416-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 432-128(pX), rX0
         MOVA   432-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   432-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   432-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   432-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   432-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   432-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   432-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   432-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 464-128(pX), rX0
         MOVA   464-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   464-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   464-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   464-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   464-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   464-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   464-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   464-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 480-128(pX), rX0
         MOVA   480-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   480-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   480-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   480-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   480-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   480-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   480-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   480-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 496-128(pX), rX0
         MOVA   496-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   496-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   496-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   496-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   496-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   496-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   496-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   496-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 528-128(pX), rX0
         MOVA   528-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   528-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   528-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   528-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   528-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   528-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   528-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   528-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 544-128(pX), rX0
         MOVA   544-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   544-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   544-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   544-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   544-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   544-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   544-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   544-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 560-128(pX), rX0
         MOVA   560-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   560-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   560-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   560-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   560-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   560-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   560-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   560-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 592-128(pX), rX0
         MOVA   592-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   592-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   592-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   592-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   592-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   592-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   592-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   592-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 608-128(pX), rX0
         MOVA   608-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   608-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   608-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   608-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   608-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   608-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   608-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   608-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 624-128(pX), rX0
         MOVA   624-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   624-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   624-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   624-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   624-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   624-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   624-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   624-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 656-128(pX), rX0
         MOVA   656-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   656-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   656-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   656-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   656-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   656-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   656-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   656-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 672-128(pX), rX0
         MOVA   672-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   672-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   672-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   672-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   672-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   672-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   672-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   672-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 688-128(pX), rX0
         MOVA   688-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   688-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   688-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   688-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   688-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   688-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   688-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   688-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 720-128(pX), rX0
         MOVA   720-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   720-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   720-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   720-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   720-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   720-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   720-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   720-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 736-128(pX), rX0
         MOVA   736-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   736-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   736-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   736-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   736-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   736-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   736-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   736-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         movapd 752-128(pX), rX0
         MOVA   752-128(pA0), rA0
         mulpd rX0, rA0
         addpd rA0, rY0

         MOVA   752-128(pA0,lda), rA0
         mulpd rX0, rA0
         addpd rA0, rY1
         MOVA   752-128(pA0,lda,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY2
         MOVA   752-128(pA0,lda3), rA0
         mulpd rX0, rA0
         addpd rA0, rY3
         MOVA   752-128(pA0,lda,4), rA0
         mulpd rX0, rA0
         addpd rA0, rY4
         MOVA   752-128(pA0,lda5), rA0
         mulpd rX0, rA0
         addpd rA0, rY5
         MOVA   752-128(pA0,lda3,2), rA0
         mulpd rX0, rA0
         addpd rA0, rY6
         MOVA   752-128(pA0,lda7), rA0
         mulpd rX0, rA0
         addpd rA0, rY7

         sub incAXm, pX
         sub incAXm, pA0
      sub incII, II
      jnz LOOPM

      cmp $0, Mr
      jz  MCLEANED

      mov Mr, II
      LOOPMCU:
         movsd -128(pX), rX0
         movsd -128(pA0), rA0
         mulsd rX0, rA0
         addsd rA0, rY0
         movsd -128(pA0,lda), rA0
         mulsd rX0, rA0
         addsd rA0, rY1
         movsd -128(pA0,lda,2), rA0
         mulsd rX0, rA0
         addsd rA0, rY2
         movsd -128(pA0,lda3), rA0
         mulsd rX0, rA0
         addsd rA0, rY3
         movsd -128(pA0,lda,4), rA0
         mulsd rX0, rA0
         addsd rA0, rY4
         movsd -128(pA0,lda5), rA0
         mulsd rX0, rA0
         addsd rA0, rY5
         movsd -128(pA0,lda3,2), rA0
         mulsd rX0, rA0
         addsd rA0, rY6
         movsd -128(pA0,lda7), rA0
         mulsd rX0, rA0
         addsd rA0, rY7
         add $4, pX
         add $4, pA0
      dec II
      jnz LOOPMCU

MCLEANED:
      haddps rY1, rY0
      haddps rY3, rY2
      haddps rY2, rY0
      movaps rY0, 0(pY)
      haddps rY5, rY4
      haddps rY7, rY6
      haddps rY6, rY4
      movaps rY4, 16(pY)
      prefY(8*4+PFYDIST(pY))
      add $8*4, pY
      add incAn, pA0
      mov pX0, pX
      mov M, II
   sub $8, N
   jnz LOOPN
/*
 * EPILOGUE: restore registers and return
 */
   movq -8(%rsp), %rbp
   movq -16(%rsp), %rbx
   movq -24(%rsp), %r12
   movq -32(%rsp), %r13
   movq -40(%rsp), %r14
   movq -48(%rsp), %r15
   ret
